/* INFOH302 - TP04-05 - Exercice 4: Composite */

import java.util.Stack;
import java.util.NoSuchElementException;


/** Itérateur in-order renvoyant chaque noeud sous forme de String, 
	ainsi que des parenthèses pour que la concaténation des Strings
	reflête la structure de l'arbre. */
public class FasterExpressionIterator implements ExpressionIterator {
	private Expression exp;
	private Stack stkAncestors;
	private int iLeftParen = 0; // Gère les parenthèses
	private int iRightParen = 0;

	public FasterExpressionIterator(Expression exp) {
		this.exp = exp;
		stkAncestors = new Stack();
		
		// Initialize stack to left-most node
		stkAncestors.push(exp);
		leftMost();
	}

	public boolean hasNext() {
		return iRightParen>0 || !stkAncestors.empty();
	}
	
	/* Go to left-most child of current node. */
	private void leftMost() {
		Expression cur = (Expression) stkAncestors.peek();
		while((cur = cur.getLeft()) != null) {
			stkAncestors.push(cur);
			iLeftParen++;
		}
	}

	public String next() {
		if(iLeftParen>0) {
			iLeftParen--;
			return "(";
		}
		if(iRightParen>0) {
			iRightParen--;
			return ")";
		}
		if(stkAncestors.empty()) {
			throw new NoSuchElementException();
		}

		Expression next = (Expression) stkAncestors.peek();
		
		// Do something really smart to update the stack
		if(next.getRight()!=null) { 
			// About to return an operator --> expand right sub-tree
			stkAncestors.push(next.getRight());
			leftMost();
		} else {
			// About to return a number --> climb up the tree
			Expression cur = (Expression) stkAncestors.pop();
			Expression parent = (Expression) stkAncestors.peek();
			while(parent.getRight() == cur) {
				iRightParen++;
				stkAncestors.pop();
				if(stkAncestors.empty()) break;
				cur = parent;
				parent = (Expression) stkAncestors.peek();
			}
		}
		return next.atomString();
	}
	
	public static void main(String[] args) {
		// ((1+2)*(4-3))/2 == 1.5 ?
		Expression ex1 = new Number(1);
		Expression ex2 = new Number(2);
		Expression ex3 = new Number(3);
		Expression ex4 = new Number(4);
		Expression plus = new PlusOperation(ex1, ex2);
		Expression minus = new MinusOperation(ex4, ex3);
		Expression mul = new MultiplyOperation(plus, minus);
		Expression div = new DivideOperation(mul, ex2);
		System.out.println(div.evaluate());
		StringBuffer sb = new StringBuffer();
		for(ExpressionIterator it = div.iterator(); it.hasNext();) {
			sb.append(it.next());
		}
		System.out.println(sb);
	}
}
